DECLARE FUNCTION CenterString(WorkString AS STRING, xWidth AS INTEGER) AS STRING
DECLARE SUB GetScreenSize()
DECLARE SUB LoadSettings()
DECLARE SUB ShowHelp()
DECLARE SUB RefreshScreen()
DECLARE SUB AddHistory(UrlToAdd AS STRING)
DECLARE SUB HistoryBack()
DECLARE SUB QuitProgram(QuitMessage AS STRING = "")
DECLARE SUB ThrowMsg(MessageContent AS STRING, SleepTime AS INTEGER = 0)
DECLARE FUNCTION TypeNewURL(ActualUrl AS STRING) AS STRING
DECLARE SUB KeybFlush()
DECLARE FUNCTION RepeatPrint(RepeatString AS STRING, RepeatAction AS INTEGER) AS STRING
DECLARE FUNCTION IsGopherUrlValid(UrlToValidate AS STRING) AS BYTE
DECLARE SUB ClearContentSpace()
DECLARE SUB DrawTUI()
DECLARE FUNCTION SelectFileToSave() AS STRING

CONST pVer AS STRING = "0.8"
CONST pDate AS STRING = "2009"

REDIM SHARED FileBuffer(1 TO 1024) AS UBYTE
DIM SHARED UrlHistory(0 TO 255) AS STRING
DIM SHARED FileSize AS INTEGER
DIM SHARED AS INTEGER xSize, ySize, OrgCursorPosX, OrgCursorPosY
DIM SHARED AS STRING CurURL, NewUrl, SaveTo

TYPE ColorStyle
  TitleBarFg AS BYTE
  TitleBarBg AS BYTE
  StatusBarFg AS BYTE
  StatusBarBg AS BYTE
  UrlBarFg AS BYTE
  UrlBarBrightFg AS BYTE
  UrlBarBg AS BYTE
  BorderFg AS BYTE
  BorderBg AS BYTE
  MsgWindowFg AS BYTE
  MsgWindowBg AS BYTE
  MsgWindowBorder AS BYTE
  DocumentFg AS BYTE
  DocumentBg AS BYTE
  GopherSelectedLinkFg AS BYTE
  GopherSelectedLinkBg AS BYTE
  GopherItemType AS BYTE
  GopherSelectedItemType AS BYTE
  GopherLink AS BYTE
  GopherBackground AS BYTE
  GopherInfo AS BYTE
  GopherError AS BYTE
END TYPE
DIM SHARED ColorScheme AS ColorStyle
DIM SHARED BoxChar(1 TO 6) AS STRING
DIM SHARED SymbolChar(1 TO 3) AS STRING
#IFDEF __FB_LINUX__    ' UTF8 encoding
  BoxChar(1) = "╔" : BoxChar(2) = "═" : BoxChar(3) = "╗"
  BoxChar(4) = "║" : BoxChar(5) = "╚" : BoxCHar(6) = "╝"
  SymbolChar(1) = "↑" : SymbolChar(2) = "░" : SymbolChar(3) = "↓"
#ENDIF
#IFDEF __FB_DOS__      ' CP 437
  BoxChar(1) = CHR(201) : BoxChar(2) = CHR(205) : BoxChar(3) = CHR(187)
  BoxChar(4) = CHR(186) : BoxChar(5) = CHR(200) : BoxCHar(6) = CHR(188)
  SymbolChar(1) = CHR(24) : SymbolChar(2) = CHR(176) : SymbolChar(3) = CHR(25)
#ENDIF
#IFDEF __FB_WIN32__    ' CP 437
  BoxChar(1) = CHR(201) : BoxChar(2) = CHR(205) : BoxChar(3) = CHR(187)
  BoxChar(4) = CHR(186) : BoxChar(5) = CHR(200) : BoxCHar(6) = CHR(188)
  SymbolChar(1) = CHR(24) : SymbolChar(2) = CHR(176) : SymbolChar(3) = CHR(25)
#ENDIF
  ' 7 bit ASCII:
  'BoxChar(1) = "+" : BoxChar(2) = "-" : BoxChar(3) = "+"
  'BoxChar(4) = "|" : BoxChar(5) = "+" : BoxCHar(6) = "+"
  'SymbolChar(1) = "#" : SymbolChar(2) = "*" : SymbolChar(3) = "#"

#INCLUDE ONCE "readtxt.bi"    ' Handler for TXT files
#INCLUDE ONCE "readmenu.bi"   ' Handler for gopher menus
#INCLUDE ONCE "loadfile.bi"   ' File loader
#INCLUDE ONCE "gophurl.bi"    ' Provide the "AnalyseURL()" function

DIM AS INTEGER FirstLoop = 1

PRINT "Digger Dwarf v"; pVer; " Copyright (C) Mateusz Viste "; pDate
SLEEP 250
OrgCursorPosX = POS(0)
OrgCursorPosY = CSRLIN

CurURL = COMMAND(1)

GetScreenSize
IF xSize < 40 AND ySize < 20 THEN PRINT "The terminal is too small. The minimal required size is 40 columns x 20 rows."

#IFDEF __FB_DOS__ ' Double-buffering possible only on the DOS version.
  SCREEN , 1, 0  ' Setting page 1 as "active", and page 0 as "display"
  PCOPY 0, 2     ' Page 0: Displayed page ; Page 1: Work page ; Page 2: Saved initial page
#ENDIF

COLOR 7, 0
CLS
LoadSettings()
DrawTUI()

WHILE IsGopherUrlValid(CurURL) = 0  ' Wrong URL, specify one!
  IF FirstLoop = 0 THEN ThrowMsg("Wrong URL! Please specify a valid URL and try again.", 1500)
  FirstLoop = 0
  ClearContentSpace()
  CurURL = TypeNewURL(CurURL)
WEND

DO
  NewUrl = AnalyseURL(CurURL, "server") & "/"
  IF VAL(AnalyseURL(CurURL, "port")) <> 70 THEN NewURL += ":" + AnalyseURL(CurURL, "port")
  NewUrl += AnalyseURL(CurURL, "GopherType") + AnalyseURL(CurURL, "selector")
  IF IsGopherUrlValid(NewURL) <> 0 THEN CurURL = NewURL
  NewURL = ""

  DrawTUI()
  SELECT CASE AnalyseURL(CurURL, "GopherType")
    CASE "0", "1", "2"
      REM do nothing, the content will be displayed on-screen
    CASE ELSE
      REM Ask where to save the file
      SaveTo = SelectFileToSave()
  END SELECT
  RefreshScreen()
  LoadFile(AnalyseURL(CurURL, "server"), VAL(AnalyseURL(CurURL, "port")), AnalyseURL(CurURL, "selector"))
  RefreshScreen()

  SELECT CASE AnalyseURL(CurURL, "GopherType")
    CASE "0"       ' Text file
      ReadTXT()
    CASE "1"       ' Gopher menu
      ReadMenu()
    CASE ELSE
      ThrowMsg("Unhandled file type, can't process. Sorry. Press any key...")
      SLEEP
      KeybFlush()
      IF LEN(UrlHistory(1)) > 0 THEN
          HistoryBack()
        ELSE
          ClearContentSpace()
          NewURL = TypeNewUrl(CurURL)
      END IF
  END SELECT

  WHILE IsGopherUrlValid(NewURL) = 0  ' Wrong URL, specify one!
    IF FirstLoop = 0 THEN ThrowMsg("Wrong URL! Please specify a valid URL and try again.", 1500)
    FirstLoop = 0
    ClearContentSpace()
    NewURL = TypeNewURL(NewURL)
  WEND

  IF UrlHistory(0) = "Do not historize" THEN  ' Do not historize if the new URL
      UrlHistory(0) = ""                      ' is already from the history.
    ELSE
      AddHistory(CurURL)
  END IF
  CurURL = NewURL
LOOP

REM  ***  END OF THE MAIN PROGRAM - SUBS AND FUNCTIONS FOLLOW  ***


FUNCTION CenterString(WorkString AS STRING, xWidth AS INTEGER) AS STRING
  DIM AS STRING Result
  IF LEN(WorkString) >= xWidth THEN
      Result = WorkString
    ELSE
      Result = SPACE((xWidth - LEN(WorkString)) \ 2) + WorkString + SPACE((xWidth - LEN(WorkString)) \ 2)
  END IF
  IF LEN(Result) < xWidth THEN Result += " "
  RETURN RESULT
END FUNCTION


SUB GetScreenSize
  DIM w AS INTEGER = Width
  xSize = LOWORD(w)
  ySize = HIWORD(w)
END SUB


SUB LoadSettings()
  REM  *** WRITE ME ***
  REM  >> Default Scheme <<
  ColorScheme.TitleBarFg = 14
  ColorScheme.TitleBarBg = 6
  ColorScheme.StatusBarFg = 0
  ColorScheme.StatusBarBg = 3
  ColorScheme.UrlBarFg = 7
  ColorScheme.UrlBarBrightFg = 15
  ColorScheme.UrlBarBg = 9
  ColorScheme.BorderFg = 3
  ColorScheme.BorderBg = 0
  ColorScheme.MsgWindowFg = 14
  ColorScheme.MsgWindowBg = 4
  ColorScheme.MsgWindowBorder = 0
  ColorScheme.DocumentFg = 7
  ColorScheme.DocumentBg = 1
  ColorScheme.GopherSelectedLinkFg = 15
  ColorScheme.GopherSelectedLinkBg = 2
  ColorScheme.GopherItemType = 8
  ColorScheme.GopherSelectedItemType = 0
  ColorScheme.GopherLink = 7
  ColorScheme.GopherBackground = 1
  ColorScheme.GopherInfo = 2
  ColorScheme.GopherError = 4
  REM  >> B/W Scheme <<
  'ColorScheme.TitleBarFg = 0
  'ColorScheme.TitleBarBg = 15
  'ColorScheme.StatusBarFg = 8
  'ColorScheme.StatusBarBg = 7
  'ColorScheme.UrlBarFg = 8
  'ColorScheme.UrlBarBrightFg = 0
  'ColorScheme.UrlBarBg = 7
  'ColorScheme.BorderFg = 7
  'ColorScheme.BorderBg = 0
  'ColorScheme.MsgWindowFg = 15
  'ColorScheme.MsgWindowBg = 7
  'ColorScheme.MsgWindowBorder = 0
  'ColorScheme.DocumentFg = 7
  'ColorScheme.DocumentBg = 0
  'ColorScheme.GopherSelectedLinkFg = 0
  'ColorScheme.GopherSelectedLinkBg = 7
  'ColorScheme.GopherItemType = 8
  'ColorScheme.GopherSelectedItemType = 8
  'ColorScheme.GopherLink = 7
  'ColorScheme.GopherBackground = 0
  'ColorScheme.GopherInfo = 7
  'ColorScheme.GopherError = 7
END SUB


SUB ShowHelp()
  REM  *** WRITE ME ***
END SUB


SUB RefreshScreen()
  #IFDEF __FB_DOS__  ' Double buffering is active
    PCOPY 1, 0       ' only on the DOS version.
  #ENDIF
END SUB


SUB AddHistory(UrlToAdd AS STRING)
  DIM x AS INTEGER
  FOR x = UBOUND(UrlHistory) TO 2 STEP -1
    UrlHistory(x) = UrlHistory(x - 1)
  NEXT x
  UrlHistory(1) = UrlToAdd
END SUB


SUB HistoryBack()
  DIM x AS INTEGER
  IF LEN(UrlHistory(1)) > 0 THEN
      NewURL = UrlHistory(1)
      UrlHistory(0) = "Do not historize"
      FOR x = 1 TO UBOUND(UrlHistory) - 1
        UrlHistory(x) = UrlHistory(x + 1)
      NEXT x
      UrlHistory(UBOUND(UrlHistory)) = ""
    ELSE
      ThrowMsg("No history available, sorry.", 1000)
      UrlHistory(0) = "Do not historize"
      NewURL = CurURL
  END IF
END SUB


SUB QuitProgram(QuitMessage AS STRING = "")
  COLOR 7, 0
  #IFDEF __FB_DOS__
    PCOPY 2, 0
    LOCATE OrgCursorPosY, OrgCursorPosX
  #ELSE
    CLS
  #ENDIF
  IF LEN(QuitMessage) > 0 THEN PRINT QuitMessage
  'DIM AS INTEGER x        '
  'PRINT                   ' Dump URL history
  'COLOR 7, 0              ' history at exit
  'PRINT "History:"        ' (usefull for debug)
  'FOR x = 1 TO 255        '
  '  IF LEN(UrlHistory(x)) > 0 THEN PRINT x & " "; UrlHistory(x)
  'NEXT x
  END
END SUB


SUB ThrowMsg(MessageContent AS STRING, SleepTime AS INTEGER = 0)
  DIM MsgArray(1 TO 12) AS STRING
  DIM WorkString AS STRING = MessageContent
  DIM AS INTEGER MsgLines = 0, MsgWidth, x, y, StartingLine, StartingColumn
  IF xSize \ 3 > 16 THEN
      MsgWidth = xSize \ 3
    ELSE
      MsgWidth = 16
  END IF
  WHILE LEN(WorkString) > 0 AND MsgLines < 12
    MsgLines += 1
    MsgArray(MsgLines) = WordWrap(WorkString, MsgWidth)
    IF LEN(MsgArray(MsgLines)) < MsgWidth THEN MsgArray(MsgLines) += SPACE(MsgWidth - LEN(MsgArray(MsgLines)))
  WEND
  LOCATE ,,0    ' Turn off the blinking cursor
  StartingLine = (ySize - MsgLines) \ 2 - 1
  StartingColumn = (xSize - MsgWidth) \ 2 - 1
  COLOR ColorScheme.MsgWindowBorder, ColorScheme.MsgWindowBg
  LOCATE StartingLine - 1, StartingColumn : PRINT BoxChar(1); RepeatPrint(BoxChar(2), MsgWidth + 2); BoxChar(3);
  FOR x = StartingLine TO StartingLine + MsgLines - 1                                         ' Draw borders
    LOCATE x, StartingColumn: PRINT BoxChar(4); SPACE(MsgWidth + 2); BoxChar(4);              ' of the message
  NEXT x                                                                                      ' window.
  LOCATE StartingLine + MsgLines, StartingColumn : PRINT BoxChar(5); RepeatPrint(BoxChar(2), MsgWidth + 2); BoxChar(6);
  COLOR ColorScheme.MsgWindowFg, ColorScheme.MsgWindowBg             '
  y = 0                                                              '
  FOR x = StartingLine TO StartingLine + MsgLines - 1                ' Fill in the message window
    y += 1                                                           ' with the message content.
    LOCATE x, StartingColumn + 2: PRINT MsgArray(y);                 '
  NEXT x                                                             '
  IF SleepTime > 0 THEN SLEEP SleepTime, 1  ' If a sleeptime has been specified, go sleep for a while
  RefreshScreen()                           ' Refresh the screen's content
  KeybFlush()                               ' Flush the keyboard buffer
END SUB


SUB KeybFlush()   ' Flush keyboard buffer
  WHILE LEN(INKEY) > 0 : WEND
END SUB


FUNCTION TypeNewURL(ActualUrl AS STRING) AS STRING
  DIM AS STRING InputBuffer, LastKey, Result
  DIM AS INTEGER MaxUrlLen
  MaxUrlLen = xSize - 15
  InputBuffer = ActualUrl
  NewURL = ""
  DO
    LOCATE 2, 13, 1    ' Locate the cursor + make sure the blinking cursor is visible
    IF LEN(InputBuffer) > MaxUrlLen THEN
        COLOR ColorScheme.UrlBarFg, ColorScheme.UrlBarBg
        PRINT "...";
        COLOR ColorScheme.UrlBarBrightFg, ColorScheme.UrlBarBg
        PRINT RIGHT(InputBuffer, MaxUrlLen - 3);
      ELSE
        COLOR ColorScheme.UrlBarBrightFg, ColorScheme.UrlBarBg
        PRINT InputBuffer;
        COLOR ColorScheme.UrlBarFg, ColorScheme.UrlBarBg
        PRINT RepeatPrint(".", MaxUrlLen - LEN(InputBuffer));
    END IF
    LOCATE 2, 13 + LEN(InputBuffer)
    KeybFlush()
    SLEEP
    LastKey = INKEY
    IF ASC(LastKey) > 31 AND LEN(LastKey) = 1 THEN InputBuffer += LastKey   ' Ignore any control codes
    IF ASC(LastKey) = 8 AND LEN(InputBuffer) > 0 THEN InputBuffer = LEFT(InputBuffer, LEN(InputBuffer) - 1)
    IF ASC(LastKey) = 9 THEN Result = CurURL
    IF ASC(LastKey) = 27 THEN QuitProgram()
    IF LastKey = CHR(13) THEN Result = InputBuffer
  LOOP UNTIL LEN(NewURL) > 0 OR LastKey = CHR(13)
  RETURN Result
END FUNCTION


FUNCTION RepeatPrint(RepeatString AS STRING, RepeatAction AS INTEGER) AS STRING
  DIM AS INTEGER x
  DIM AS STRING Result = ""
  FOR x = 1 TO RepeatAction
    Result += RepeatString
  NEXT x
  RETURN Result
END FUNCTION


FUNCTION IsGopherUrlValid(UrlToValidate AS STRING) AS BYTE
  DIM AS BYTE Result
  IF LEN(AnalyseURL(UrlToValidate, "server")) = 0 OR LEN(AnalyseURL(UrlToValidate, "port")) = 0 THEN
      Result = 0
    ELSE
      Result = 1
  END IF
  RETURN Result
END FUNCTION


SUB ClearContentSpace()
  DIM AS INTEGER x
  COLOR 0, ColorScheme.GopherBackground
  FOR x = 3 TO ySize - 1
    LOCATE x, 2 : PRINT SPACE(xSize - 2);
  NEXT x
END SUB


SUB DrawTUI()
  DIM AS INTEGER y, MaxUrlLen
  MaxUrlLen = xSize - 15
  ClearContentSpace()
  LOCATE 1, 1: COLOR ColorScheme.TitleBarFg, ColorScheme.TitleBarBg : PRINT CenterString("Digger Dwarf v" + pVer, xSize);
  COLOR ColorScheme.BorderFg, ColorScheme.BorderBg
  LOCATE 2, 1 : PRINT BoxChar(1);
  LOCATE 2, xSize : PRINT BoxChar(3);
  FOR y = 3 TO ySize - 1
    LOCATE y, 1: PRINT BoxChar(4);
    LOCATE y, xSize: PRINT BoxChar(4);
  NEXT y
  COLOR ColorScheme.UrlBarFg, ColorScheme.UrlBarBg
  LOCATE 2, 2: COLOR ColorScheme.UrlBarFg, ColorScheme.UrlBarBg : PRINT " [" + STRING(xSize - 6, ASC(".")) + "] ";
  LOCATE 2, 4: COLOR ColorScheme.UrlBarBrightFg, ColorScheme.UrlBarBg: PRINT "gopher://";
  IF LEN(CurURL) > MaxUrlLen THEN
      PRINT LEFT(CurURL, MaxUrlLen - 5); "(...)";
    ELSE
      PRINT CurURL;
  END IF
  COLOR ColorScheme.StatusBarFg, ColorScheme.StatusBarBg
  LOCATE ySize, 1: PRINT SPACE(xSize);
  LOCATE ySize, 1: PRINT " Available Keys: Up/Down, ENTER, BackSpc, TAB, F1, F10 ";
END SUB


FUNCTION SelectFileToSave() AS STRING
  DIM AS STRING Result
  REM   ***  WRITE ME!  ***
  RETURN Result
END FUNCTION